''' Serializers/Deserializers (SerDe) convert between configuration strings (or numbers) and
different Python structures.

You should subclass :class:`ConfigSerDeBase` and override the following static fields:

- ``name``: the description of the structure this serde can convert, for example "Comma-separated
  list of integers". This text will be used in the documentation generated by the ConfigBase CLI.
- ``serialize``: Take the Python structure and returns a string representation.
- ``deserialize``: Take a string representation of the value and return a Python structure.
'''

from typing import Any, List, Sequence
from abc import ABC, abstractmethod

class ConfigSerDeBase(ABC):
    ''' Defines a serializer / deserializer interface. '''

    name : str = 'Describe the structure here'

    @staticmethod
    @abstractmethod
    def serialize(value: Any) -> str:
        ''' Serializes a config value to a string.

        Args:
            value (Any): a Python object to be serialized.

        Returns:
            The object serialized into a string.
        '''

    @staticmethod
    @abstractmethod
    def deserialize(value: str) -> Any:
        ''' Deserializes a string to a config value.

        Args:
            value (str): a Python object serialized into a string

        Returns:
            The the restored Python object.
        '''


class IntegerListSerDe(ConfigSerDeBase):
    ''' De/serializes a string containing a comma-separated list of integers.

    Example string: ``0, 1, 2``
    '''

    name : str = 'Comma-separated list of integers'

    @staticmethod
    def serialize(value: Sequence[int]) -> str:
        ''' Serializes a list of integers into a string.

        Args:
            value (Sequence[int]): The list of integers.

        Returns:
            The list of integers serialized into a string.
        '''
        return ', '.join(str(e) for e in value)

    @staticmethod
    def deserialize(value: str) -> List[int]:
        '''Restores a list of integers from a string.

        Args:
            value (str): A string containing a serialized list of integers.

        Returns:
            The list of integers restored from the string.

        Raises:
            Exception: exceptions related to invalid string format.
        '''
        return [int(e) for e in value.split(',')]



''' The ``config.tool`` module defines a simple command-line interface (CLI) program that can be
used as a development tool. It creates snippets that you can copy-paste into the Panorama
application project files like graph.json, or package.json of the business logic package.
'''
import argparse, json, textwrap
from .config import ConfigBase

def cli(name: str, config: ConfigBase) -> None:
    ''' Creates a simple command line interface for printing config snippets.

    Args:
        name (str): The name of your Panorama app
        config (ConfigBase): Instance of your configuration structure, subclass of ConfigBase.
    '''
    parser = argparse.ArgumentParser(description=textwrap.dedent(
        f'''\
        Configuration snippet generator for {name} application.

        This program can generate json and markdown snippets that you can copy-paste to the
        metadata and package definitions of your AWS Panorama project. The snippets contain the
        definitions of the application parameters in the required format. The following formats
        are supported:

         - nodes: generates a json snippet to be pasted in the nodeGraph.nodes field of graph.json
         - edges: generates a json snippet to be pasted in the nodeGraph.edges field of graph.json
         - interface: generates json a snippet to be pasted in nodePackage.interfaces field of the
             package.json of the application code package
         - markdown: generates a markdown snippet that you can paste to the README of your project,
             or other parts of the documentation.
         '''),
         formatter_class=argparse.RawDescriptionHelpFormatter
    )
    function_map = {
        'nodes' : lambda config, _: print(json.dumps(config.get_panorama_definitions(), indent=4)),
        'edges' : lambda config, code_node: print(json.dumps(config.get_panorama_edges(code_node), indent=4)),
        'interface': lambda config, _: print(json.dumps(config.get_panorama_app_interface(), indent=4)),
        'markdown': lambda config, _: print(config.get_panorama_markdown_doc()),
    }

    parser.add_argument('command', choices=function_map.keys(),
        help='Prints configuration snippets for graph.json nodes, edges, application '
             'interface in package.json, or in markdown format.'
    )
    parser.add_argument('--code-node', type=str, default='code_node',
        help='Code node name (used in edges snippet)'
    )

    args = parser.parse_args()
    func = function_map[args.command]
    func(config, args.code_node)

if __name__=='__main__':
    cli('config_base', ConfigBase())



''' This module defines :class:`~backpack.config.ConfigBase`, a base class for Panorama application
configurations. The class offers two basic functionalities:

- parse parameters from the Panorama application input ports (``panoramasdk.node.inputs``)
- generate configuration file snippets for ``graph.json`` and ``package.json`` in your Panorama
  project. For more information about how to use the CLI, refer to :meth:`~backpack.config.tool`.
'''

import dataclasses
from typing import Sequence, List, Any, Type, TypeVar, Mapping
import dataclasses

T = TypeVar('T', bound='ConfigBase')

class ConfigBase:
    ''' Base class for configuration structures.

    Subclasses must be also dataclasses.
    '''

    TYPE_MAP = {
        int: 'int32',
        float: 'float32',
        str: 'string',
        bool: 'boolean'
    }

    def __init__(self) -> None:
        assert dataclasses.is_dataclass(self), 'ConfigBase instances must be also dataclasses.'

    def param_walker(self, prefix: Sequence[List]=[]):
        fields = dataclasses.fields(self)
        for f in fields:
            name_components = prefix + [f.name]
            if dataclasses.is_dataclass(f.type):
                obj = getattr(self, f.name)
                for subres in obj.param_walker(name_components):
                    yield subres
            else:
                if 'type' in f.metadata:
                    typename = f.metadata['type']
                else:
                    typename = ConfigBase.TYPE_MAP.get(f.type)
                serde = f.metadata.get('serde')
                if serde is not None and f.default is not None:
                    default = serde.serialize(f.default)
                else:
                    default = f.default
                if typename is None:
                    raise ValueError(f'Dataclass {self} field has unsupported type: {f}')
                name = '_'.join(name_components)
                doc = f.metadata.get('__doc__', f.metadata.get('doc'))
                yield (name, typename, default, doc, name_components, f)

    def get_panorama_definitions(self) -> List[Mapping[str, Any]]:
        ''' Generate the ``nodeGraph.nodes`` snippet in ``graph.json``.

        Returns:
            A list of dictionaries containing the application parameter node definitions.
        '''
        return [
            {
                'name': name,
                'interface': typename,
                'value': default,
                'overridable': True,
                'decorator': {
                    'title': name,
                    'description': doc
                }
            }
            for (name, typename, default, doc, *_) in self.param_walker()
        ]

    def get_panorama_edges(self, code_node_name) -> List[Mapping[str, str]]:
        ''' Generate the ``nodeGraph.edges`` snippet in ``graph.json``

        Returns:
            A list of dictionaries containing the application edge definitions.
        '''
        return [
            {
                "producer": name,
                "consumer": code_node_name + "." + name
            }
            for (name, *_) in self.param_walker()
        ]

    def get_panorama_app_interface(self) -> List[Mapping[str, str]]:
        ''' Generate the application interface snippet in app node ``package.json``.

        Returns:
            A list of dictionaries containing the elements of the application interface definition.
        '''
        return [
            {
                "name": name,
                "type": typename
            }
            for (name, typename, *_) in self.param_walker()
        ]

    def get_panorama_markdown_doc(self) -> str:
        ''' Generates a markdown table of the parameters that can be used in documentation.

        Returns:
            Markdown formatted text containing the parameter documentation.
        '''
        header = (
            '| name | type    | default | description |\n'
            '|------|---------|---------|-------------|\n'
        )
        body = '\n'.join([
            f'| {name} | {typename} | {default} | {doc} |'
            for name, typename, default, doc, *_ in self.param_walker()
        ])
        return header + body

    @classmethod
    def from_panorama_params(cls: Type[T], inputs: 'panoramasdk.port') -> T:
        ''' Parses the config values form AWS Panorama input parameters.

        A new Config object is created with the default values. If a particular value is
        found in the input parameter, its value will override the default value.

        Args:
            inputs (panoramasdk.port): The input port of the Panorama application node.

        Returns:
            The config instance filled with the parameter values read from the input port.
        '''
        result = cls()
        for name, typename, default, doc, name_components, f in result.param_walker():
            obj = result
            for name_part in name_components[:-1]:
                obj = getattr(obj, name_part)
            key = name_components[-1]
            if not hasattr(inputs, name):
                continue
            value = getattr(inputs, name).get()
            serde = f.metadata.get('serde')
            if serde is not None:
                value = serde.deserialize(value)
            setattr(obj, key, value)
        return result

    def asdict(self):
        return dataclasses.asdict(self)



''' This module contains code to standardize the configuration of Panorama applications via
deploy-time parameters. It also contains a CLI (command-line interface) that generates
configuration snippets for the metadata and descriptor json files of your Panorama project.
'''

from .tool import cli
from .config import ConfigBase
from .serde import ConfigSerDeBase, IntegerListSerDe



''' This module contains config serdes for objects from backpack.geometry module. '''

from typing import Optional
import dataclasses
import json

from backpack.geometry import PolyLine

from .serde import ConfigSerDeBase

class PolyLineSerDe(ConfigSerDeBase):
    ''' De/serializes a json string containing a polygon definition.

    Example string: ``[[0.1, 0.1], [0.9, 0.1], [0.9, 0.9], [0.5, 0.5], [0.1, 0.9]]``
    '''
    name : str = 'JSON encoded Polygon'

    @staticmethod
    def serialize(value: PolyLine) -> str:
        ''' Serializes a PolyLine into a string.

        Args:
            value (PolyLine): The PolyLine to be serialized.

        Returns:
            The PolyLine serialized into a string.
        '''
        return json.dumps(dataclasses.astuple(value)[0])

    @staticmethod
    def deserialize(value: str) -> Optional[PolyLine]:
        '''
        Restores a PolyLine from a string.

        Args:
            value (str): A string containing a serialized PolyLine.

        Returns:
            The PolyLine restored from the string or None if the string was empty.

        Raises:
            Exception: exceptions related to invalid string format.
        '''
        return (
            PolyLine.from_value(json.loads(value))
                if value is not None and len(value) > 0 else
            None
        )
